#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.kernel.reservation_client import ClientReservationFactory
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.reservation_list import ReservationList


class ReservationListTest(unittest.TestCase):
    def test_create(self):
        r_list = ReservationList()
        self.assertEqual(r_list.size(), 0)
        self.assertIsNotNone(r_list.rset_wrapper_list)
        self.assertIsNotNone(r_list.cycle_to_rset)
        self.assertEqual(r_list.count, 0)

    def make_reservation(self, rid: str) -> ABCReservationMixin:
        return ClientReservationFactory.create(rid=ID(uid=rid))

    def test_add_reservation(self):
        r_list = ReservationList()
        for i in range(10):
            for j in range(10):
                c = (i*10) + j
                res = self.make_reservation(str(c))
                r_list.add_reservation(reservation=res, cycle=i)
                self.assertEqual(c+1, r_list.size())

    def test_get_reservations(self):
        r_list = ReservationList()
        for i in range(10):
            for j in range(10):
                c = (i * 10) + j
                res = self.make_reservation(str(c))
                r_list.add_reservation(reservation=res, cycle=i)
                self.assertEqual(c + 1, r_list.size())

        for i in range(10):
            low = 10 * i
            high = 10 * (i + 1)

            rset = r_list.get_reservations(cycle=i)
            self.assertIsNotNone(rset)
            self.assertEqual(10, rset.size())

            for r in rset.values():
                value = int(str(r.get_reservation_id()))
                self.assertTrue(low <= value < high)

        rset = r_list.get_reservations(cycle=10)
        self.assertIsNotNone(rset)
        self.assertEqual(0, rset.size())

    def test_tick(self):
        r_list = ReservationList()
        for i in range(10):
            for j in range(10):
                c = (i*10) + j
                res = self.make_reservation(str(c))
                r_list.add_reservation(reservation=res, cycle=i)
                self.assertEqual(c+1, r_list.size())
                self.assertEqual(c+1, len(r_list.reservation_id_to_cycle))

        for i in range(10):
            rset = r_list.get_reservations(cycle=i)
            self.assertIsNotNone(rset)
            self.assertEqual(10, rset.size())

            expected_size = (10 - i) * 10
            self.assertEqual(expected_size, r_list.size())
            r_list.tick(cycle=i)

            rset = r_list.get_reservations(cycle=i)

            self.assertIsNotNone(rset)
            self.assertEqual(0, rset.size())

        self.assertEqual(0, r_list.size())
        self.assertEqual(0, len(r_list.reservation_id_to_cycle))

    def check_exists(self, holdings: ReservationList, reservation: ABCReservationMixin, cycle: int):
        rset = holdings.get_reservations(cycle=cycle)
        return rset.contains(reservation=reservation)

    def test_remove_reservation(self):
        r_list = ReservationList()
        for i in range(10):
            for j in range(10):
                c = (i*10) + j
                res = self.make_reservation(str(c))
                r_list.add_reservation(reservation=res, cycle=i)
                self.assertEqual(c+1, r_list.size())
                self.assertEqual(c+1, len(r_list.reservation_id_to_cycle))

        for i in range(10):
            for j in range(10):
                c = (i * 10) + j
                res = self.make_reservation(str(c))
                exist = self.check_exists(r_list, res, i)
                self.assertTrue(exist)
                r_list.remove_reservation(reservation=res)
                exist = self.check_exists(holdings=r_list, reservation=res, cycle=i)
                self.assertTrue(not exist)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import enum
from typing import Tuple, Union, List

import requests
from fim.user import GraphFormat

@enum.unique
class Status(enum.Enum):
    OK = 1
    INVALID_ARGUMENTS = 2
    FAILURE = 3


class OrchestratorHelper:
    HTTP_OK = 200
    PROP_DATA = "data"
    PROP_SLIVERS = "slivers"
    PROP_SLICES = "slices"

    def __init__(self):
        self.host = "http://localhost:8700"
        self.token = "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiIsImtpZCI6ImI0MTUxNjcyMTExOTFlMmUwNWIyMmI1NGIxZDNiNzY2N2U3NjRhNzQ3NzIyMTg1ZTcyMmU1MmUxNDZmZTQzYWEifQ.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.bjSrZgGQDXM4g50wcw-sWEJVMR5g614Od6Sav0Im9fg8MsU68Jpi3uJ8wEkGR8cVO6_FsaYsAhJPDSSuElaeN8bUrkRuCUpx6BoaQ-G7J_JmlAQvH-GIgmYhu9T9epBQLUf_wKFVrbsemEPPzAKofJ58hTchtHTlRWt0hJ7m7eEeBirXJDQeKOSyRFKJ5R61to69tqlf5G_QxuqsaWRcIHdgGxkPlNKNW2aV2aKVvwyde2rbr647NFyXzACpj4ir1aS8WBoTBEhVAa6NV8V9r3QQq6OqKLktj65rRKX4q6DyhEl01QiFBg--JLAAZ2NaOntY4DC8yP_6VnfbvM_8zsYBKdZMgii6Wf0-UNXNEyK47ZfXTCa5v7Tp0ozTNc0_Hd3TFRdNqQ3kCicVZFa671mapBC43lD-pYRloTASUADa89L2VAxxRCl89-i5Gc01le-6ZiUy0moLT_COD8qEQvixnOD4vnb_7VMSiCVSNLzkGM-qCtdMGeYC9PF3y35np1aEkttVGg_mKljsvDp7NMMBcGFP9YJCWVNEQZTvGqvm-HX5iQvXzTXr7jdcxljko1jZH2lHlD9hlNeLazhm6caAiGsSITMXjpzM7rbXfPVf1mEIqIha3NxrQnKDK3vAsWVnFNrOg74LWCwVXZGQggR9yg2XHpV9dBpOscECHHQ"
        self.headers = {
            'accept': 'application/json',
            'Authorization': f"Bearer {self.token}"
        }
        self.headers_with_content = {
            'accept': 'application/json',
            'Authorization': f"Bearer {self.token}",
            'Content-Type': "text/plain"
        }
        self.ssh_key = "empty ssh string"

    def resources(self, level: int = 1):
        url = f"{self.host}/resources?level={level}&force_refresh=true"
        return requests.get(url, headers=self.headers, verify=False)

    def portal_resources(self, graph_format: GraphFormat = GraphFormat.JSON_NODELINK):
        url = f"{self.host}/portalresources?graph_format={graph_format.name}"
        headers = {'accept': 'application/json'}
        return requests.get(url, headers=headers, verify=False)

    def create(self, slice_graph: str, slice_name: str,
               lease_end_time: str = None) -> Tuple[Status, Union[list, requests.Response]]:
        if lease_end_time is None:
            url = f"{self.host}/slices/create?name={slice_name}&ssh_key={self.ssh_key}"
        else:
            url = f"{self.host}/slices/create?name={slice_name}&ssh_key={self.ssh_key}&lease_end_time={lease_end_time}"
        response = requests.post(url, headers=self.headers_with_content, verify=False, data=slice_graph)
        if response.status_code == self.HTTP_OK:
            reservations = response.json()[self.PROP_DATA]
            return Status.OK, reservations
        else:
            return Status.FAILURE, response

    def delete(self, slice_id: str) -> Tuple[Status, requests.Response]:
        url = f"{self.host}/slices/delete/{slice_id}"
        response = requests.delete(url, headers=self.headers, verify=False)
        if response.status_code == self.HTTP_OK:
            return Status.OK, response
        else:
            return Status.FAILURE, response

    def slices(self, slice_id: str = None,
               state: str = "Active") -> Tuple[Status, Union[list, str, requests.Response]]:
        url = f"{self.host}/slices?state={state}"
        if slice_id is not None:
            url = f"{self.host}/slices/{slice_id}?graph_format=GRAPHML"

        response = requests.get(url, headers=self.headers, verify=False)
        if response.status_code == self.HTTP_OK:
            slices = response.json()[self.PROP_DATA]
            return Status.OK, slices
        else:
            return Status.FAILURE, response

    def slivers(self, slice_id: str, sliver_id: str = None) -> Tuple[Status, Union[list, str, requests.Response]]:
        url = f"{self.host}/slivers?slice_id={slice_id}"
        if sliver_id is not None:
            url = f"{self.host}/slivers/{sliver_id}?slice_id={slice_id}"

        response = requests.get(url, headers=self.headers, verify=False)
        if response.status_code == self.HTTP_OK:
            reservations = response.json()[self.PROP_DATA]
            return Status.OK, reservations
        else:
            return Status.FAILURE, response

    def renew(self, slice_id: str, new_lease_end_time: str) -> Tuple[Status, Union[requests.Response, None]]:
        """
        Renew a slice
        @param token fabric token
        @param slice_id slice_id
        @param new_lease_end_time new_lease_end_time
        @return Tuple containing Status and List of Reservation Id failed to extend
        """
        response = None
        try:
            # Set the tokens
            url = f"{self.host}/slices/renew/{slice_id}?new_lease_end_time={new_lease_end_time}"
            response = requests.post(url, headers=self.headers, verify=False)
            if response.status_code != self.HTTP_OK:
                return Status.FAILURE, response

            return Status.OK, response
        except Exception as e:
            return Status.FAILURE, response


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod, ABC
from datetime import datetime
from typing import TYPE_CHECKING, List, Union, Tuple, Dict

from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.kernel.slice import SliceTypes

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_broker_proxy import ABCBrokerProxy
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.plugins.handlers.configuration_mapping import ConfigurationMapping
    from fabric_cf.actor.core.container.maintenance import Site
    from fabric_cf.actor.core.kernel.poa import Poa


class ABCDatabase(ABC):
    """
    IDatabase is the base database layer interface. It specifies methods for managing slices, reservations,
    broker proxies, and configuration mapping files.
    """

    @abstractmethod
    def actor_added(self, *, actor):
        """
        Performs initialization actions as a result of the actor being
        added to the container.

        @throws Exception in case of error
        """

    @abstractmethod
    def add_broker(self, *, broker: ABCBrokerProxy):
        """
        Adds a new broker proxy record.

        @param broker broker proxy

        @throws Exception in case of error
        """

    @abstractmethod
    def add_reservation(self, *, reservation: ABCReservationMixin):
        """
        Adds a new record to the database representing this reservation
        object.

        @param reservation reservation

        @throws Exception in case of error
        """

    @abstractmethod
    def add_slice(self, *, slice_object: ABCSlice):
        """
        Adds a new record to the database representing this slice
        object.

        @param slice_object Slice object

        @throws Exception in case of error
        """

    @abstractmethod
    def remove_broker(self, *, broker: ABCBrokerProxy):
        """
        Removes the specified broker proxy record.

        @param broker broker proxy

        @throws Exception in case of error
        """

    @abstractmethod
    def remove_reservation(self, *, rid: ID):
        """
        Removes the corresponding reservation object.

        @param rid reservation id

        @throws Exception in case of error
        """

    @abstractmethod
    def remove_slice(self, *, slice_id: ID):
        """
        Removes the corresponding database slice record.

        @param slice_id slice name

        @throws Exception in case of error
        """

    @abstractmethod
    def set_actor_name(self, *, name: str):
        """
        Sets the name of the actor this database belongs to.

        @param name actor name
        """

    @abstractmethod
    def update_broker(self, *, broker: ABCBrokerProxy):
        """
        Updates the specified broker proxy record.

        @param broker broker proxy

        @throws Exception in case of error
        """

    @abstractmethod
    def update_reservation(self, *, reservation: ABCReservationMixin):
        """
        Updates the corresponding reservation object.

        @param reservation reservation

        @throws Exception in case of error
        """

    @abstractmethod
    def update_slice(self, *, slice_object: ABCSlice):
        """
        Updates the corresponding database slice record.

        @param slice_object slice object

        @throws Exception in case of error
        """

    @abstractmethod
    def get_reservations(self, *, slice_id: ID = None, graph_node_id: str = None, project_id: str = None,
                         email: str = None, oidc_sub: str = None, rid: ID = None, states: list[int] = None,
                         site: str = None, rsv_type: list[str] = None, start: datetime = None,
                         end: datetime = None, ip_subnet: str = None, host: str = None) -> List[ABCReservationMixin]:
        """
        Retrieves the reservations.

        @return list of reservations

        @throws Exception in case of error
        """

    @abstractmethod
    def get_components(self, *, node_id: str, states: list[int], rsv_type: list[str], component: str = None,
                       bdf: str = None, start: datetime = None, end: datetime = None,
                       excludes: List[str] = None) -> Dict[str, List[str]]:
        """
        Returns components matching the search criteria
        @param node_id: Worker Node ID to which components belong
        @param states: list of states used to find reservations
        @param rsv_type: type of reservations
        @param component: component name
        @param bdf: Component's PCI address
        @param start: start time
        @param end: end time
        @param excludes: Excludes the list of reservations

        NOTE# For P4 switches; node_id=node+renc-p4-sw  component=ip+192.168.11.8 bdf=p1

        @return Dictionary with component name as the key and value as list of associated PCI addresses in use.
        """

    @abstractmethod
    def get_client_reservations(self, *, slice_id: ID = None) -> List[ABCReservationMixin]:
        """
        Retrieves the client reservations

        @return list of reservations

        @throws Exception in case of error
        """

    @abstractmethod
    def get_slices(self, *, slice_id: ID = None, slice_name: str = None, project_id: str = None, email: str = None,
                   states: list[int] = None, oidc_sub: str = None, slc_type: List[SliceTypes] = None,
                   limit: int = None, offset: int = None, lease_end: datetime = None,
                   search: str = None, exact_match: bool = False) -> List[ABCSlice] or None:
        """
        Retrieves the specified slices.

        @param slice_id slice id
        @param slice_name slice name
        @param project_id project id
        @param email email
        @param states states
        @param oidc_sub oidc sub
        @param slc_type slice type
        @param limit limit
        @param offset offset
        @param lease_end lease_end
        @param search: search term applied
        @param exact_match: Exact Match for Search term

        @return list of slices

        @throws Exception in case of error
        """

    @abstractmethod
    def get_slice_count(self, *, project_id: str = None, email: str = None, states: list[int] = None,
                        oidc_sub: str = None, slc_type: List[SliceTypes] = None,
                        excluded_projects: List[str] = None) -> int:
        """
        Retrieves the slices count.

        @param project_id project id
        @param email email
        @param states states
        @param oidc_sub oidc sub
        @param slc_type slice type
        @param excluded_projects excluded_projects

        @return number of slices matching the filter criteria

        @throws Exception in case of error
        """

    @abstractmethod
    def increment_metrics(self, *, project_id: str, oidc_sub: str, slice_count: int = 1) -> bool:
        """
        Add or Update Metrics

        @param project_id project id
        @param oidc_sub oidc sub
        @param slice_count slice_count

        @return true or false

        @throws Exception in case of error
        """

    @abstractmethod
    def get_metrics(self, *, project_id: str, oidc_sub: str, excluded_projects: List[str] = None) -> list:
        """
        Get Metrics
        @param project_id: project id
        @param oidc_sub: user id
        @param excluded_projects: list of project ids to exclude

        @return list of metrics
        """

    @abstractmethod
    def initialize(self):
        """
        Initializes the object.

        @throws Exception if initialization fails
        """

    @abstractmethod
    def get_holdings(self, *, slice_id: ID = None) -> List[ABCReservationMixin]:
        """
        Retrieves all reservations representing resources held by this
        actor Broker/Controller.

        @return list of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_broker_reservations(self) -> List[ABCReservationMixin]:
        """
        Retrieves all reservations for which this actor acts as a
        broker.

        @return list of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_authority_reservations(self) -> List[ABCReservationMixin]:
        """
        Retrieves all reservations for which this actor acts as a site.

        @return list of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_reservations_by_rids(self, *, rid: List[str]) -> List[ABCReservationMixin]:
        """
        Retrieves the specified reservation records.
        The order in the return vector is the same order as @rids
        @param rid rids
        @return list of reservations
        @throws Exception in case of error
        """

    @abstractmethod
    def add_delegation(self, *, delegation: ABCDelegation):
        """
        Add delegation
        @params delegation: delegation
        """

    @abstractmethod
    def update_delegation(self, *, delegation: ABCDelegation):
        """
        Update delegation
        @params delegation: delegation
        """

    @abstractmethod
    def remove_delegation(self, *, dlg_graph_id: str):
        """
        Remove delegation
        @params dlg_graph_id: dlg_graph_id
        """

    @abstractmethod
    def get_delegation(self, *, dlg_graph_id: str) -> ABCDelegation or None:
        """
        Get Delegation
        @params dlg_graph_id: dlg_graph_id
        @return Delegation
        """

    @abstractmethod
    def get_delegations(self, *, slice_id: ID = None, states: List[int] = None) -> List[ABCDelegation]:
        """
        Get delegations
        @params slice_id: slice_id
        @params state: state
        @return Delegations
        """

    @abstractmethod
    def add_config_mapping(self, key: str, config_mapping: ConfigurationMapping):
        """
        Add configuration Mapping
        @param key key
        @param config_mapping config mapping
        """

    @abstractmethod
    def get_config_mappings(self) -> List[ConfigurationMapping]:
        """
        Return all config mappings
        """

    @abstractmethod
    def get_brokers(self) -> List[ABCBrokerProxy] or None:
        """
        Return all brokers
        """

    @abstractmethod
    def add_site(self, *, site: Site):
        """
        Add site
        @param site site
        """

    def update_site(self, *, site: Site):
        """
        Update Site
        @param site site
        """
    def remove_site(self, *, site_name: str):
        """
        Remove Site
        @param site_name site name
        """
    def get_site(self, *, site_name: str) -> Site or None:
        """
        Get site
        @param site_name site name
        @return Site
        """

    def get_sites(self) -> List[Site]:
        """
        Return all sites
        @return list of sites
        """

    def add_maintenance_properties(self, *, properties: dict):
        """
        Add maintenance properties
        @param properties properties
        """

    def update_maintenance_properties(self, *, properties: dict):
        """
        Update maintenance properties
        @param properties properties
        """

    def remove_maintenance_properties(self):
        """
        Remove maintenance properties
        """

    def get_maintenance_properties(self) -> dict:
        """
        Get maintenance Properties
        @return properties
        """

    @abstractmethod
    def add_poa(self, *, poa: Poa):
        """
        Adds a new record to the database representing this slice
        object.

        @param poa POA object

        @throws Exception in case of error
        """

    @abstractmethod
    def update_poa(self, *, poa: Poa):
        """
        Updates the corresponding database poa record.

        @param poa poa info object

        @throws Exception in case of error
        """

    @abstractmethod
    def remove_poa(self, *, poa_id: str):
        """
        Removes the corresponding database slice record.

        @param poa_id poa id

        @throws Exception in case of error
        """

    @abstractmethod
    def get_poas(self, *, poa_id: str = None, email: str = None, sliver_id: ID = None, slice_id: ID = None,
                 project_id: str = None, limit: int = None, offset: int = None, last_update_time: datetime = None,
                 states: list[int] = None) -> Union[List[Poa] or None]:
        """
        Get POAs
        @param poa_id poa id
        @param email email
        @param sliver_id sliver id
        @param slice_id slice id
        @param project_id project id
        @param limit limit
        @param offset offset
        @param states states
        @param last_update_time last update time
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest

from fabric_cf.actor.core.time.actor_clock import ActorClock


class ActorClockTest(unittest.TestCase):
    def _get_clock(self, offset: int, length: int):
        return ActorClock(beginning_of_time=offset, cycle_millis=length)

    def test_create(self):
        clock = self._get_clock(0, 1)
        self.assertEqual(0, clock.beginning_of_time)
        self.assertEqual(1, clock.cycle_millis)
        self.assertEqual(0, clock.get_beginning_of_time())
        self.assertEqual(1, clock.get_cycle_millis())

        clock = self._get_clock(1000, 10)
        self.assertEqual(1000, clock.beginning_of_time)
        self.assertEqual(10, clock.cycle_millis)
        self.assertEqual(1000, clock.get_beginning_of_time())
        self.assertEqual(10, clock.get_cycle_millis())

    def test_cycle(self):
        offset = 1000
        length = 10
        clock = self._get_clock(1000, 10)

        ms = offset
        for i in range(100):
            exp = i
            self.assertEqual(exp, clock.cycle(when=ActorClock.from_milliseconds(milli_seconds=ms)))
            ms += length

        ms = offset
        for i in range(100):
            exp = i
            for j in range(length):
                self.assertEqual(exp, clock.cycle(when=ActorClock.from_milliseconds(milli_seconds=ms)))
                ms += 1

    def test_date(self):
        offset = 1000
        length = 10
        clock = self._get_clock(1000, 10)

        ms = offset
        for i in range(100):
            self.assertEqual(ActorClock.from_milliseconds(milli_seconds=ms), clock.date(cycle=i))
            ms += length

    def test_cycle_start_end_date(self):
        offset = 1000
        length = 10
        clock = self._get_clock(1000, 10)

        start = offset
        end = offset + length - 1

        for i in range(100):
            self.assertEqual(ActorClock.from_milliseconds(milli_seconds=start), clock.cycle_start_date(cycle=i))
            self.assertEqual(ActorClock.from_milliseconds(milli_seconds=end), clock.cycle_end_date(cycle=i))
            self.assertEqual(start, clock.cycle_start_in_millis(cycle=i))
            self.assertEqual(end, clock.cycle_end_in_millis(cycle=i))
            start += length
            end += length

    def test_misc(self):
        length = 10
        clock = self._get_clock(1000, 10)

        ms = 0
        for i in range(100):
            temp = clock.get_millis(cycle=i)
            self.assertEqual(ms, temp)
            self.assertEqual(i, clock.convert_millis(millis=temp))
            ms += length


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)


class Notice:
    def __init__(self):
        self.notice_string = None

    def get_notice(self):
        """
        Get notice string
        :return:
        """
        return self.notice_string

    def add(self, *, msg: str, ex: Exception = None):
        """
        Addd message
        :param msg: message
        :param ex: exception
        :return:
        """
        if msg is None:
            return

        if self.notice_string is None:
            self.notice_string = msg
        else:
            self.notice_string += msg

        if ex is not None:
            self.notice_string += str(ex)

    def clear(self):
        """
        Clear the notice
        :return:
        """
        self.notice_string = None

    def is_empty(self) -> bool:
        """
        Returns true if empty; false otherwise
        :return: true if empty; false otherwise
        """
        return self.notice_string is None

    def __str__(self):
        if self.notice_string is not None:
            return self.notice_string
        return ""



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric_cf.actor.boot.configuration import ActorConfig
from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.exceptions import ActorException
from fabric_cf.actor.core.plugins.handlers.config_token import ConfigToken
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.reservation_set import ReservationSet
from fabric_cf.actor.core.kernel.resource_set import ResourceSet


class Policy(ABCPolicy):
    """
    Base class for all policy implementations.
    """
    def __init__(self, *, actor: ABCActorMixin = None):
        """
        Creates a new instance.
        @params actor : actor this policy belongs to
        """
        self.initialized = False
        if actor is not None:
            self.actor = actor
            self.logger = actor.get_logger()
            self.clock = actor.get_actor_clock()
        else:
            self.actor = None
            self.logger = None
            self.clock = None
            self.guid = ID()
        self.guid = ID()
        self.properties = None

    def close(self, *, reservation: ABCReservationMixin):
        """
        Close a reservation
        @params reservation: reservation about to be closed
        """

    def closed(self, *, reservation: ABCReservationMixin):
        """
        Notifies the policy that a reservation has been closed.
        @params reservation: reservation about to be closed
        """

    def configuration_complete(self, *, action: str, token: ConfigToken, out_properties: dict):
        """
        Notifies the policy that a configuration action for the object
        represented by the token parameter has completed.

        @params action : configuration action. See Config.Target*
        @params token : object or a token for the object whose configuration action has completed
        @params outProperties : output properties produced by the configuration action
        """

    def error(self, *, message: str):
        """
        Logs the specified error and throws an exception.
        @params message: error message
        @raises Exception in case of error
        """
        self.logger.error(message)
        raise ActorException(message)

    def extend(self, *, reservation: ABCReservationMixin, resources: ResourceSet, term: Term):
        return

    def finish(self, *, cycle: int):
        return

    def get_closing(self, *, cycle: int) -> ReservationSet:
        return None

    def get_guid(self) -> ID:
        return self.guid

    def initialize(self, *, config: ActorConfig):
        """
        Initialize the policy object
        """
        if not self.initialized:
            if self.actor is None:
                raise ActorException("Missing actor")

            if self.logger is None:
                self.logger = self.actor.get_logger()

            if self.logger is None:
                raise ActorException("Missing logger")

            if self.clock is None:
                self.clock = self.actor.get_actor_clock()

            if self.clock is None:
                raise ActorException("Missing clock")

            if config and config.get_policy() and config.get_policy().get_properties():
                self.set_properties(properties=config.get_policy().get_properties())

            self.initialized = True

    def internal_error(self, *, message: str):
        """
        Logs the specified error and throws an exception.
        @params message: error message
        @raises Exception in case of error
        """
        self.logger.error("Internal error: " + message)
        raise ActorException("Internal error: " + message)

    def log_error(self, *, message: str):
        """
        logs an error
        @params message: error message
        """
        self.logger.error("Internal mapper error: " + message)

    def log_warn(self, *, message: str):
        """
        logs a warning
        @params message: warning message
        """
        self.logger.warning("Internal mapper warning: " + message)

    def prepare(self, *, cycle: int):
        return

    def query(self, *, p):
        return None

    def remove(self, *, reservation: ABCReservationMixin):
        return

    def reset(self):
        return

    def recovery_starting(self):
        return

    def revisit(self, *, reservation: ABCReservationMixin):
        return

    def revisit_delegation(self, *, delegation: ABCDelegation):
        return

    def recovery_ended(self):
        return

    def set_actor(self, *, actor: ABCActorMixin):
        self.actor = actor
        if actor is not None:
            self.logger = actor.get_logger()

    def closed_delegation(self, *, delegation: ABCDelegation):
        return

    def set_logger(self, logger):
        self.logger = logger

    def set_properties(self, properties: dict):
        self.properties = properties


from fabric_cf.orchestrator.swagger_server.models.metrics import Metrics  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response import metrics_controller as rc


def metrics_overview_get(excluded_projects=None):  # noqa: E501
    """Control Framework metrics overview

    Control Framework  metrics overview # noqa: E501

    :param excluded_projects: List of projects to exclude from the metrics overview
    :type excluded_projects: List[str]

    :rtype: Metrics
    """
    return rc.metrics_overview_get(excluded_projects=excluded_projects)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime, timezone
from typing import List, Dict, Tuple, Union

from fim.slivers.maintenance_mode import MaintenanceInfo, MaintenanceState

from fabric_cf.actor.core.apis.abc_database import ABCDatabase
from fabric_cf.actor.core.common.constants import Constants


class Site:
    def __init__(self, *, name: str, maint_info: MaintenanceInfo):
        """
        Represents a Site in maintenance
        """
        self.name = name
        self.maintenance_info = maint_info
        if self.maintenance_info is not None:
            self.maintenance_info.finalize()
        # Contains allowed projects/users
        self.properties = {}

    def get_name(self) -> str:
        return self.name

    def get_maintenance_info(self) -> MaintenanceInfo:
        return self.maintenance_info

    def set_properties(self, *, properties: dict):
        self.properties = properties

    def get_properties(self) -> dict:
        return self.properties

    def is_in_maintenance(self) -> bool:
        if self.maintenance_info is not None:
            now = datetime.now(timezone.utc)
            site_info = self.maintenance_info.get(self.name)
            if site_info is not None and (site_info.state == MaintenanceState.Maint or
                                         (site_info.state == MaintenanceState.PreMaint and site_info.deadline <= now)):
                return True
        return False

    def get_state(self) -> MaintenanceState:
        site_info = self.maintenance_info.get(self.name)
        if site_info is not None:
            return site_info.state
        else:
            self.maintenance_info.finalize()
            for name, entry in self.maintenance_info.iter():
                return entry.state

    def is_worker_in_maintenance(self, *, worker: str) -> bool:
        # Whole site is in Maintenance
        if self.is_in_maintenance():
            return True

        # Check if the specific worker is in Maintenance
        entry = self.maintenance_info.get(name=worker)
        now = datetime.now(timezone.utc)

        if entry is not None and (entry.state == MaintenanceState.Maint or
                                  (entry.state == MaintenanceState.PreMaint and entry.deadline <= now)):
            return True

        return False

    def __str__(self):
        return f"Name: {self.name} MaintInfo: {self.maintenance_info} Properties: {self.properties}"

    def clone_maintenance_info(self) -> Union[MaintenanceInfo or None]:
        if self.maintenance_info is not None:
            return self.maintenance_info.copy()
        return None

    def update_maintenance_info(self, maint_info: MaintenanceInfo):
        self.maintenance_info = maint_info


class Maintenance:
    @staticmethod
    def update_maintenance_mode(*, database: ABCDatabase, properties: Dict[str, str], sites: List[Site] = None):
        """
        Update Maintenance Mode at Testbed/Site/Worker Level
        - Tesbed level Maintenance - single Site object is passed with Name = ALL
        - Site level Maintenance - single Site object per site is passed with Name = SiteName
        - Worker level Maintenance - single Site object per site with one entry per worker
        @param database database
        @param properties properties container project ids/ user emails
        @param sites Maintenance information for the sites
        """
        for s in sites:
            # Set the list of allowed projects/users at the site level
            if properties is not None:
                s.set_properties(properties=properties)

            # Get Current Maintenance mode for the Site
            existing_site = database.get_site(site_name=s.get_name())
            # Site entry exists
            if existing_site is not None:
                # Update the Properties {project id/user email information)
                if properties is not None:
                    existing_site.set_properties(properties=properties)
                # Site level Maintenance Update
                if s.get_maintenance_info().get(s.get_name()) is not None:
                    database.update_site(site=s)
                # Worker level Maintenance Update
                else:
                    new_maint_info = existing_site.clone_maintenance_info()
                    if new_maint_info.get(s.get_name()):
                        new_maint_info.rem(s.get_name())
                    for worker_name, entry in s.get_maintenance_info().list_details():
                        # Remove existing entry
                        if new_maint_info.get(worker_name):
                            new_maint_info.rem(worker_name)

                        # Add worker entry using the new information only if worker is in Maintenance
                        if entry.state != MaintenanceState.Active:
                            new_maint_info.add(worker_name, entry)
                    new_maint_info.finalize()
                    existing_site.update_maintenance_info(maint_info=new_maint_info)
                    database.update_site(site=existing_site)
            # Adding Maintenance State First Time
            else:
                database.add_site(site=s)

    @staticmethod
    def is_testbed_in_maintenance(*, database: ABCDatabase) -> Tuple[bool, Dict[str, str] or None]:
        test_bed = database.get_site(site_name=Constants.ALL)
        if test_bed is not None:
            return test_bed.is_in_maintenance(), test_bed.get_properties()

        return False, None

    @staticmethod
    def is_site_in_maintenance(*, database: ABCDatabase, site_name: str) -> Tuple[bool, Site or None]:
        site = database.get_site(site_name=site_name)
        if site is None:
            return False, None

        return site.is_in_maintenance(), site

    @staticmethod
    def is_sliver_provisioning_allowed(*, database: ABCDatabase, project: str, email: str, site: str,
                                       worker: str) -> Tuple[bool, str or None]:
        """
        Determine if sliver can be provisioned
        Sliver provisioning can be prohibited if Testbed or Site or Worker is in maintenance mode
        Sliver provisioning in maintenance mode may be allowed for specific projects/users
        @param database database
        @param project project
        @param email user's email
        @param site site name
        @param worker worker name
        @return True if allowed; False otherwise
        """
        status, site = Maintenance.is_site_in_maintenance(database=database, site_name=site)

        if not status and site is None:
            return True, None

        projects = site.get_properties().get(Constants.PROJECT_ID)
        users = site.get_properties().get(Constants.USERS)

        if project is not None and projects is not None and project in projects:
            return True, None

        if email is not None and users is not None and email in users:
            return True, None

        if status:
            return False, f"Site {site.get_name()} in {site.get_state()}"

        if worker is not None and site.is_worker_in_maintenance(worker=worker):
            return False, f"Worker {worker} on {site.get_name()} in {site.get_state()}"

        return True, None

    @staticmethod
    def is_slice_provisioning_allowed(*, database: ABCDatabase, project: str, email: str) -> bool:
        """
        Determine if slice can be provisioned
        Slice provisioning can be prohibited if Testbed is in maintenance mode
        Slice provisioning in maintenance mode may be allowed for specific projects/users
        @param database database
        @param project project
        @param email user's email
        @return True if allowed; False otherwise
        """

        status, properties = Maintenance.is_testbed_in_maintenance(database=database)

        if not status:
            return True

        users = properties.get(Constants.USERS)
        projects = properties.get(Constants.PROJECT_ID)

        if project is not None and projects is not None and project in projects:
            return True

        if users is not None and email is not None and email in users:
            return True

        return False



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import traceback
from datetime import datetime, timezone

from confluent_kafka.cimpl import Producer
from fim.graph.abc_property_graph import GraphFormat

from fabric_cf.actor.core.apis.abc_broker_mixin import ABCBrokerMixin
from fabric_cf.actor.core.apis.abc_timer_task import ABCTimerTask
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.broker_policy import BrokerPolicy


class BrokerQueryModelPublisher(ABCTimerTask):
    """
    This class queries Broker for BQM and publishes it to Kafka Topic
    Queries are done periodically based on the configured timer and published to configured kafka topic.
    Published information will be used by the Portal
    """
    def __init__(self, *, broker: ABCBrokerMixin, logger, producer: Producer, kafka_topic: str):
        self.broker = broker
        self.logger = logger
        self.producer = producer
        self.kafka_topic = kafka_topic

    def execute(self):
        """
        Process a claim timeout
        """
        try:
            request = BrokerPolicy.get_broker_query_model_query(level=1, bqm_format=GraphFormat.JSON_NODELINK)
            response = self.broker.query(query=request, caller=self.broker.get_identity())
            if response is None:
                self.logger.error(f"Could not get broker query model!")
                return

            status = response.get(Constants.QUERY_RESPONSE_STATUS, None)
            bqm = response.get(Constants.BROKER_QUERY_MODEL, None)

            if status is None or status == 'False' or bqm is None or bqm == '':
                self.logger.error(f"Could not get broker query model!")
                return

            key = f'{self.broker.get_guid()}-{datetime.now(timezone.utc).timestamp()}'

            # Publish to Kafka
            if self.producer is not None and self.kafka_topic is not None:
                self.producer.produce(topic=self.kafka_topic, key=key, value=bqm)
            else:
                self.logger.error(f"Could not get broker query model")
        except Exception as e:
            self.logger.debug(f"Exception occurred in BQM publisher: {e}")
            self.logger.error(traceback.format_exc())




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import List

from fabric_mb.message_bus.messages.close_delegations_avro import CloseDelegationsAvro
from fabric_mb.message_bus.messages.close_reservations_avro import CloseReservationsAvro
from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.get_delegations_avro import GetDelegationsAvro
from fabric_mb.message_bus.messages.get_sites_request_avro import GetSitesRequestAvro
from fabric_mb.message_bus.messages.maintenance_request_avro import MaintenanceRequestAvro
from fabric_mb.message_bus.messages.poa_info_avro import PoaInfoAvro
from fabric_mb.message_bus.messages.remove_delegation_avro import RemoveDelegationAvro
from fabric_mb.message_bus.messages.reservation_state_avro import ReservationStateAvro
from fabric_mb.message_bus.messages.get_reservations_state_request_avro import GetReservationsStateRequestAvro
from fabric_mb.message_bus.messages.add_slice_avro import AddSliceAvro
from fabric_mb.message_bus.messages.get_reservations_request_avro import GetReservationsRequestAvro
from fabric_mb.message_bus.messages.get_slices_request_avro import GetSlicesRequestAvro
from fabric_mb.message_bus.messages.remove_reservation_avro import RemoveReservationAvro
from fabric_mb.message_bus.messages.remove_slice_avro import RemoveSliceAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.site_avro import SiteAvro
from fabric_mb.message_bus.messages.slice_avro import SliceAvro
from fabric_mb.message_bus.messages.update_reservation_avro import UpdateReservationAvro
from fabric_mb.message_bus.messages.update_slice_avro import UpdateSliceAvro

from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.apis.abc_mgmt_actor import ABCMgmtActor
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.manage.kafka.kafka_proxy import KafkaProxy


class KafkaActor(KafkaProxy, ABCMgmtActor):
    def get_guid(self) -> ID:
        return self.management_id

    def prepare(self, *, callback_topic: str):
        self.callback_topic = callback_topic

    def toggle_maintenance_mode(self, actor_guid: str, callback_topic: str, sites: List[SiteAvro] = None,
                                projects: str = None, users: str = None):
        props = {}
        if projects is not None:
            props[Constants.PROJECT_ID] = projects
        if users is not None:
            props[Constants.USERS] = users
        request = MaintenanceRequestAvro(properties=props, actor_guid=actor_guid,
                                         callback_topic=callback_topic, sites=sites)

        status, response = self.send_request(request)

        return response.status.code == 0

    def get_slices(self, *, slice_id: ID = None, slice_name: str = None, email: str = None, project: str = None,
                   states: List[int] = None, limit: int = None, offset: int = None,
                   user_id: str = None, search: str = None, exact_match: bool = False) -> List[SliceAvro] or None:
        request = GetSlicesRequestAvro()
        request = self.fill_request_by_id_message(request=request, email=email, slice_id=slice_id,
                                                  slice_name=slice_name, states=states)
        status, response = self.send_request(request)
        if response is not None:
            return response.slices

        return None

    def remove_slice(self, *, slice_id: ID) -> bool:
        request = RemoveSliceAvro()
        request = self.fill_request_by_id_message(request=request, slice_id=slice_id)
        status, response = self.send_request(request)

        return status.code == 0

    def add_slice(self, *, slice_obj: SliceAvro) -> ID:
        ret_val = None
        request = AddSliceAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.callback_topic = self.callback_topic
        request.message_id = str(ID())
        request.slice_obj = slice_obj

        status, response = self.send_request(request)

        if status.code == 0:
            ret_val = ID(uid=response.get_result())

        return ret_val

    def update_slice(self, *, slice_obj: SliceAvro, modify_state: bool = False) -> bool:
        request = UpdateSliceAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.callback_topic = self.callback_topic
        request.message_id = str(ID())
        request.slice_obj = slice_obj

        status, response = self.send_request(request)

        return status.code == 0

    def accept_update_slice(self, *, slice_id: ID) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def delete_slice(self, *, slice_id: ID) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def get_reservations(self, *, states: List[int] = None, slice_id: ID = None,
                         rid: ID = None, oidc_claim_sub: str = None, email: str = None, rid_list: List[str] = None,
                         type: str = None, site: str = None, node_id: str = None,
                         host: str = None, ip_subnet: str = None) -> List[ReservationMng]:
        request = GetReservationsRequestAvro()
        request = self.fill_request_by_id_message(request=request, slice_id=slice_id,
                                                  states=states, email=email, rid=rid,
                                                  type=type, site=site, host=host, ip_subnet=ip_subnet)
        status, response = self.send_request(request)

        if status.code == 0:
            return response.reservations

    def get_sites(self, *, site: str) -> List[SiteAvro] or None:
        request = GetSitesRequestAvro()
        request = self.fill_request_by_id_message(request=request, site=site)
        status, response = self.send_request(request)

        if status.code == 0:
            return response.sites
        return None

    def get_delegations(self, *, slice_id: ID = None, states: List[int] = None,
                        delegation_id: str = None) -> List[DelegationAvro]:
        request = GetDelegationsAvro()
        request = self.fill_request_by_id_message(request=request, slice_id=slice_id,
                                                  states=states, delegation_id=delegation_id)
        status, response = self.send_request(request)

        if status.code == 0:
            return response.delegations

    def remove_reservation(self, *, rid: ID) -> bool:
        request = RemoveReservationAvro()
        request = self.fill_request_by_id_message(request=request, rid=rid)
        status, response = self.send_request(request)

        return status.code == 0

    def close_reservation(self, *, rid: ID) -> bool:
        request = CloseReservationsAvro()
        request = self.fill_request_by_id_message(request=request, rid=rid)
        status, response = self.send_request(request)

        return status.code == 0

    def close_reservations(self, *, slice_id: ID) -> bool:
        request = CloseReservationsAvro()
        request = self.fill_request_by_id_message(request=request, slice_id=slice_id)
        status, response = self.send_request(request)

        return status.code == 0

    def update_reservation(self, *, reservation: ReservationMng) -> bool:
        request = UpdateReservationAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.callback_topic = self.callback_topic
        request.message_id = str(ID())
        request.reservation = reservation

        status, response = self.send_request(request)

        return status.code == 0

    def get_reservation_state_for_reservations(self, *, reservation_list: List[str]) -> List[ReservationStateAvro]:
        request = GetReservationsStateRequestAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.callback_topic = self.callback_topic
        request.message_id = str(ID())
        request.reservation_ids = []
        for r in reservation_list:
            request.reservation_ids.append(str(r))

        status, response = self.send_request(request)

        if status.code == 0:
            return response.reservation_states

    def clone(self):
        return KafkaActor(guid=self.management_id,
                          kafka_topic=self.kafka_topic,
                          auth=self.auth, logger=self.logger,
                          message_processor=self.message_processor,
                          producer=self.producer)

    def get_name(self) -> str:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def remove_delegation(self, *, did: str) -> bool:
        request = RemoveDelegationAvro()
        request = self.fill_request_by_id_message(request=request, delegation_id=did)
        status, response = self.send_request(request)

        return status.code == 0

    def close_delegation(self, *, did: str) -> bool:
        request = CloseDelegationsAvro()
        request = self.fill_request_by_id_message(request=request, delegation_id=did)
        status, response = self.send_request(request)

        return status.code == 0

    def get_poas(self, *, states: List[int] = None, slice_id: ID = None, rid: ID = None,
                 email: str = None, poa_id: str = None, project_id: str = None,
                 limit: int = 200, offset: int = 0) -> List[PoaInfoAvro]:
        raise NotImplemented



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_authority_proxy import ABCAuthorityProxy
    from fabric_cf.actor.core.apis.abc_broker_proxy import ABCBrokerProxy
    from fabric_cf.actor.core.apis.abc_client_callback_proxy import ABCClientCallbackProxy
    from fabric_cf.actor.core.apis.abc_client_policy import ABCClientPolicy
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.kernel.resource_set import ResourceSet
    from fabric_cf.actor.core.util.resource_type import ResourceType


class ABCClientReservation(ABCReservationMixin):
    """
    IClientReservation defines the reservation interface for actors acting as clients of other actors.
    """
    @abstractmethod
    def get_authority(self) -> ABCAuthorityProxy:
        """
        Returns the authority who issued the lease for this reservation.
        @returns site authority or null
        """

    @abstractmethod
    def get_broker(self) -> ABCBrokerProxy:
        """
        Returns a proxy to the broker linked to this reservation. Can be None
        @returns broker linked to this reservation. Can be None
        """

    @abstractmethod
    def get_previous_ticket_term(self) -> Term:
        """
        Returns the previous ticket term.
        @returns previous ticket term
        """

    @abstractmethod
    def get_renew_time(self) -> int:
        """
        Returns the cached reservation renewal time. Used during recovery
        @returns cached reservation renewal time
        """

    @abstractmethod
    def get_suggested_resources(self) -> ResourceSet:
        """
        Returns the resources suggested to the policy for a new/extend request for the reservation.
        @returns suggested resources
        """

    @abstractmethod
    def get_suggested_term(self) -> Term:
        """
        Returns the term suggested to the policy for a new/extend request for the reservation.
        @returns suggested term
        """

    @abstractmethod
    def get_suggested_type(self) -> ResourceType:
        """
        Returns the most recently suggest resource type.
        @returns suggested type
        """

    @abstractmethod
    def get_ticket_sequence_in(self) -> int:
        """
        Returns the reservation sequence number for incoming ticket/extend ticket messages.
        @returns reservation sequence number for incoming ticket/extend ticket messages
        """

    @abstractmethod
    def get_ticket_sequence_out(self) -> int:
        """
        Returns the reservation sequence number for outgoing ticket/extend ticket messages.
        @returns reservation sequence number for outgoing ticket/extend ticket messages
        """

    @abstractmethod
    def get_ticket_term(self) -> Term:
        """
        Returns the current ticket term. Note that getTerm will return the currently active term.
        This can be either the ticket term or the lease term.
        @returns current ticket term
        """

    @abstractmethod
    def is_exported(self) -> bool:
        """
        Checks if the reservation represents exported resources.
        @returns true if the reservation represents exported resources
        """

    @abstractmethod
    def is_renewable(self) -> bool:
        """
        Checks if the reservation is renewable.
        @returns true if the reservation is renewable
        """

    @abstractmethod
    def set_broker(self, *, broker: ABCBrokerProxy) -> bool:
        """
        Sets the broker who will issue tickets for the reservation. This
        method can be called only for reservations in the Nascent state.
        @params broker broker request tickets from
        @raises Exception if the reservation is in the wrong state
        """

    @abstractmethod
    def set_exported(self, *, exported: bool):
        """
        Sets the exported flag.
        @params exported: flag value
        """

    @abstractmethod
    def set_renewable(self, *, renewable: bool):
        """
        Sets the renewable flag.
        @params renewable: flag value
        """

    @abstractmethod
    def get_renewable(self) -> bool:
        """
        Gets the renewable flag.
        """

    @abstractmethod
    def set_renew_time(self, *, time: int):
        """
        Caches the reservation renewal time. This information is used to simplify recovery.
        @params time: reservation renewal time
        """

    @abstractmethod
    def set_suggested(self, *, term: Term, resources: ResourceSet):
        """
        Sets the term and resources suggested to the policy for a new/extend request for the reservation.
        @params term : suggested term
        @params resources : suggested resources
        """

    @abstractmethod
    def set_suggested_term(self, *, term: Term):
        """
        Sets the term suggested to the policy for a new/extend request for the reservation.
        @params term : suggested term
        """

    @abstractmethod
    def set_suggested_resources(self, *, resources: ResourceSet):
        """
        Sets the resources suggested to the policy for a new/extend request for the reservation.
        @params resources : suggested resources
        """

    @abstractmethod
    def set_ticket_sequence_in(self, *, sequence: int):
        """
        Sets the reservation sequence number for incoming ticket/extend ticket messages.
        @params sequence : sequence number
        """

    @abstractmethod
    def set_ticket_sequence_out(self, *, sequence: int):
        """
        Sets the reservation sequence number for outgoing ticket/extend ticket messages.
        @params sequence : sequence number
        """

    @abstractmethod
    def get_update_notices(self) -> str:
        """
        Returns a string describing the reservation status.
        @returns status string
        """

    @abstractmethod
    def set_policy(self, *, policy: ABCClientPolicy):
        """
        Sets the policy associated with this reservation.
        @params policy : policy
        """

    @abstractmethod
    def get_client_callback_proxy(self) -> ABCClientCallbackProxy:
        """
        Returns the client callback proxy for this reservation.
        @returns IClientCallbackProxy
        """

    @abstractmethod
    def get_notices_dict(self) -> str:
        """
        Return Notices as JSON String
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import queue
import threading
import traceback
from typing import List

from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.iterable_queue import IterableQueue
from fabric_cf.orchestrator.core.active_status_checker import ActiveStatusChecker
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.i_status_update_callback import IStatusUpdateCallback
from fabric_cf.orchestrator.core.status_checker import StatusChecker, Status
from fabric_cf.orchestrator.core.watch_entry import WatchEntry


class ReservationStatusUpdateThread:
    """
    This thread allows expressing interest in completion of certain reservations and can run callbacks on other specified
    reservations when the status changes accordingly

    The periodic thread polls reservations of interest to determine whether the they have reached the required state,
    upon which the callback is invoked
    """
    MODIFY_CHECK_PERIOD = 5 # seconds

    def __init__(self):
        self.controller = None
        self.thread_lock = threading.Lock()
        self.reservation_lock = threading.Lock()
        self.active_watch = queue.Queue()
        self.stopped_worker = threading.Event()

        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

        self.thread = None
        self.stopped = False

    def start(self):
        """
        Start
        :return:
        """
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise OrchestratorException("This ReservationStatusUpdateThread has already been started")

            self.thread = threading.Thread(target=self.periodic)
            self.thread.setName(self.__class__.__name__)
            self.thread.setDaemon(True)
            self.thread.start()

        finally:
            self.thread_lock.release()

    def stop(self):
        """
        Stop
        :return:
        """
        self.stopped = True
        self.stopped_worker.set()
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the ReservationStatusUpdateThread is running. Interrupting it")
                try:
                    temp.join()
                except Exception as e:
                    self.logger.error("Could not join ReservationStatusUpdateThread thread {}".format(e))
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def periodic(self):
        """
        Periodic
        :return:
        """
        self.logger.debug(f"Reservation Status Update Thread started")
        while not self.stopped_worker.wait(timeout=self.MODIFY_CHECK_PERIOD):
            self.run()
        self.logger.debug(f"Reservation Status Update Thread exited")

    def __add_active_status_watch(self, *, we: WatchEntry):
        try:
            self.reservation_lock.acquire()
            self.logger.debug(f"Added watch entry {we}")
            self.active_watch.put_nowait(we)
        finally:
            self.reservation_lock.release()

    def add_active_status_watch(self, *, watch: ID, callback: IStatusUpdateCallback):
        """
        Watch for transition to Active or Failed. Callback is called when reservation has either Failed or became Active
        @param watch - reservation to watch
        @param callback - callback object
        """
        if watch is None or callback is None:
            self.logger.info(f"watch {watch} callback {callback}, ignoring")
            return

        self.__add_active_status_watch(we=WatchEntry(watch=watch, callback=callback))

    def process_watch_list(self, *, watch_list: List[WatchEntry], watch_type: str, status_checker: StatusChecker):
        """
        Process watch list
        :param watch_list:
        :param watch_type:
        :param status_checker:
        :return:
        """
        to_remove = []
        self.logger.debug(f"Scanning {watch_type} watch list {len(watch_list)}")

        if self.controller is None:
            from fabric_cf.orchestrator.core.orchestrator_kernel import OrchestratorKernelSingleton
            self.controller = OrchestratorKernelSingleton.get().get_management_actor()

        for watch_entry in watch_list:
            status, reservation = status_checker.check(controller=self.controller, rid=watch_entry.reservation_to_watch)
            self.logger.info(f"Status------- {status} for {reservation.get_reservation_id()}")
            if status == Status.OK:
                to_remove.append(watch_entry)
                watch_entry.callback.success(controller=self.controller, reservation=reservation)
            elif status == Status.NOT_OK:
                to_remove.append(watch_entry)
                watch_entry.callback.failure(controller=self.controller, reservation=reservation)

        self.logger.debug(f"Removing {watch_type} entries from watch {len(to_remove)}")

        for we in to_remove:
            watch_list.remove(we)

        # Add the Not Ready entries back
        for we in watch_list:
            self.__add_active_status_watch(we=we)

    def run(self):
        # wake up periodically and check the status of reservations (state or unit modify properties)
        # and call appropriate callbacks if necessary.
        # scan both lists and check if any of the
        # reservation groups on them are ready for callbacks
        # remove those ready for callbacks off the lists and invoke callbacks
        # outside the critical section
        # NOTE: because callback can create another callback, which can add watches,
        # we must make this code re-entrant hence the need to change the watch list
        # and only then call the callbacks

        try:
            active_watch_list = []

            try:
                self.reservation_lock.acquire()
                if not self.active_watch.empty():
                    for a in IterableQueue(source_queue=self.active_watch):
                        active_watch_list.append(a)
            finally:
                self.reservation_lock.release()

            self.process_watch_list(watch_list=active_watch_list, watch_type="active",
                                    status_checker=ActiveStatusChecker())

        except Exception as e:
            self.logger.error(f"RuntimeException: {e} continuing")
            self.logger.error(traceback.format_exc())



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest
import time
from fabric_cf.actor.core.apis.abc_database import ABCDatabase
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.plugins.substrate.db.substrate_actor_database import SubstrateActorDatabase
from fabric_cf.actor.test.base_test_case import BaseTestCase


class SubstrateTestBase(BaseTestCase, unittest.TestCase):
    from fabric_cf.actor.core.container.globals import Globals
    Globals.config_file = "./config/config.test.yaml"
    Constants.SUPERBLOCK_LOCATION = './state_recovery.lock'

    from fabric_cf.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    def make_actor_database(self) -> ABCDatabase:
        db = SubstrateActorDatabase(user=self.db_user, password=self.db_pwd, database=self.db_name, db_host=self.db_host,
                                    logger=self.logger)
        return db

    def test_actor(self):
        actor = self.get_actor()
        self.assertIsNotNone(actor)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from datetime import datetime
from typing import TYPE_CHECKING, List

from fabric_mb.message_bus.messages.lease_reservation_avro import LeaseReservationAvro
from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.reservation_predecessor_avro import ReservationPredecessorAvro
from fabric_mb.message_bus.messages.result_delegation_avro import ResultDelegationAvro
from fabric_mb.message_bus.messages.result_broker_query_model_avro import ResultBrokerQueryModelAvro
from fabric_mb.message_bus.messages.result_proxy_avro import ResultProxyAvro
from fabric_mb.message_bus.messages.result_string_avro import ResultStringAvro
from fabric_mb.message_bus.messages.result_strings_avro import ResultStringsAvro
from fabric_mb.message_bus.messages.result_avro import ResultAvro
from fim.slivers.base_sliver import BaseSliver
from fim.user import GraphFormat

from fabric_cf.actor.core.apis.abc_actor_runnable import ABCActorRunnable
from fabric_cf.actor.core.apis.abc_controller_reservation import ABCControllerReservation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants, ErrorCodes
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.kernel.reservation_client import ClientReservationFactory
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.manage.converter import Converter
from fabric_cf.actor.core.manage.management_object import ManagementObject
from fabric_cf.actor.core.manage.management_utils import ManagementUtils
from fabric_cf.actor.core.proxies.kafka.translate import Translate
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.apis.abc_client_actor_management_object import ABCClientActorManagementObject
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.core.broker_policy import BrokerPolicy

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
    from fabric_mb.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
    from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
    from fabric_cf.actor.core.apis.abc_client_actor import ABCClientActor
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin


class ClientActorManagementObjectHelper(ABCClientActorManagementObject):
    def __init__(self, *, client: ABCClientActor):
        self.client = client
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

    def get_brokers(self, *, caller: AuthToken, broker_id: ID = None) -> ResultProxyAvro:
        result = ResultProxyAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:
            brokers = None
            if broker_id is None:
                brokers = self.client.get_brokers()
            else:
                broker = self.client.get_broker(guid=broker_id)
                if broker is not None:
                    brokers = [broker]
                    result.proxies = Converter.fill_proxies(proxies=brokers)
                else:
                    result.status.set_code(ErrorCodes.ErrorNoSuchBroker.value)
                    result.status.set_message(ErrorCodes.ErrorNoSuchBroker.interpret())
            if brokers is not None:
                result.proxies = Converter.fill_proxies(proxies=brokers)
        except Exception as e:
            self.logger.error("get_brokers {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def add_broker(self, *, broker: ProxyAvro, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()

        if broker is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:
            proxy = Converter.get_agent_proxy(mng=broker)
            if proxy is None:
                result.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            else:
                self.client.add_broker(broker=proxy)
        except Exception as e:
            self.logger.error("add_broker {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def update_broker(self, *, broker: ProxyAvro, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()

        if broker is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:
            proxy = Converter.get_agent_proxy(mng=broker)
            if proxy is None:
                result.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            else:
                self.client.update_broker(broker=proxy)
        except Exception as e:
            self.logger.error("update_broker {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def get_broker_query_model(self, *, broker: ID, caller: AuthToken, id_token: str,
                               level: int, graph_format: GraphFormat, start: datetime = None,
                               end: datetime = None, includes: str = None,
                               excludes: str = None) -> ResultBrokerQueryModelAvro:
        result = ResultBrokerQueryModelAvro()
        result.status = ResultAvro()

        if broker is None or caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:

            b = self.client.get_broker(guid=broker)
            if b is not None:
                request = BrokerPolicy.get_broker_query_model_query(level=level, bqm_format=graph_format,
                                                                    start=start, end=end, includes=includes,
                                                                    excludes=excludes)
                response = ManagementUtils.query(actor=self.client, actor_proxy=b, query=request)
                result.model = Translate.translate_to_broker_query_model(query_response=response, level=level)
            else:
                result.status.set_code(ErrorCodes.ErrorNoSuchBroker.value)
                result.status.set_message(ErrorCodes.ErrorNoSuchBroker.interpret())
        except Exception as e:
            self.logger.error("get_broker_query_model {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def add_reservation_private(self, *, reservation: TicketReservationAvro):
        result = ResultAvro()
        slice_id = ID(uid=reservation.get_slice_id())
        rset = Converter.get_resource_set(res_mng=reservation)
        term = Term(start=ActorClock.from_milliseconds(milli_seconds=reservation.get_start()),
                    end=ActorClock.from_milliseconds(milli_seconds=reservation.get_end()))

        broker = None

        if reservation.get_broker() is not None:
            broker = ID(uid=reservation.get_broker())

        rid = None
        if reservation.get_reservation_id() is not None:
            rid = ID(uid=reservation.get_reservation_id())
        else:
            rid = ID()
        rc = ClientReservationFactory.create(rid=rid, resources=rset, term=term)
        rc.set_renewable(renewable=reservation.is_renewable())

        if rc.get_state() != ReservationStates.Nascent or rc.get_pending_state() != ReservationPendingStates.None_:
            result.set_code(ErrorCodes.ErrorInvalidReservation.value)
            result.set_message("Only reservations in Nascent.None can be added")
            return None, result

        slice_obj = self.client.get_slice(slice_id=slice_id)

        if slice_obj is None:
            result.set_code(ErrorCodes.ErrorNoSuchSlice.value)
            result.set_message(ErrorCodes.ErrorNoSuchSlice.interpret())
            return None, result

        rc.set_slice(slice_object=slice_obj)

        proxy = None

        if broker is None:
            proxy = self.client.get_default_broker()
        else:
            proxy = self.client.get_broker(guid=broker)

        if proxy is None:
            result.set_code(ErrorCodes.ErrorNoSuchBroker.value)
            result.set_message(ErrorCodes.ErrorNoSuchBroker.interpret())
            return None, result

        rc.set_broker(broker=proxy)
        self.client.register(reservation=rc)
        return rc.get_reservation_id(), result

    def add_reservation(self, *, reservation: TicketReservationAvro, caller: AuthToken) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()

        if reservation is None or reservation.get_slice_id() is None or caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:
            class Runner(ABCActorRunnable):
                def __init__(self, *, parent):
                    self.parent = parent

                def run(self):
                    return self.parent.add_reservation_private(reservation=reservation)

            self.client.execute_on_actor_thread(runnable=Runner(parent=self))
        except Exception as e:
            self.logger.error("add_reservation {}".format(e))
            self.logger.error(traceback.format_exc())
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def add_reservations(self, *, reservations: List[TicketReservationAvro], caller: AuthToken) -> ResultStringsAvro:
        result = ResultStringsAvro()
        result.status = ResultAvro()

        if reservations is None or caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        for r in reservations:
            if r.get_slice_id() is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
                return result

        try:
            class Runner(ABCActorRunnable):
                def __init__(self, *, parent):
                    self.parent = parent

                def run(self):
                    result = []
                    try:
                        for r in reservations:
                            rr, status = self.parent.add_reservation_private(reservation=r)
                            if rr is not None:
                                result.append(str(rr))
                            else:
                                raise ManageException("Could not add reservation")
                    except Exception:
                        for r in reservations:
                            self.parent.client.unregister(reservation=r)
                        result.clear()

                    return result

            rids, result.status = self.client.execute_on_actor_thread_and_wait(runnable=Runner(parent=self))

            if result.status.get_code() == 0:
                for r in rids:
                    result.result.append(r)
        except Exception as e:
            self.logger.error("add_reservations {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def demand_reservation_rid(self, *, rid: ID, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()

        if rid is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:
            class Runner(ABCActorRunnable):
                def __init__(self, *, actor: ABCClientActor):
                    self.actor = actor

                def run(self):
                    self.actor.demand(rid=rid)
                    return None

            self.client.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.client))
        except Exception as e:
            self.logger.error("demand_reservation_rid {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def demand_reservation(self, *, reservation: ReservationMng, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()

        if reservation is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:
            class Runner(ABCActorRunnable):
                def __init__(self, *, actor: ABCClientActor, logger):
                    self.actor = actor
                    self.logger = logger

                def __add_predecessors(self, predecessors: List[ReservationPredecessorAvro],
                                       res: ABCControllerReservation):
                    for pred in predecessors:
                        if pred.get_reservation_id() is None:
                            self.logger.warning(f"Predecessor specified for rid={res.get_reservation_id()} "
                                                "but missing reservation id of predecessor")
                            continue

                        predid = ID(uid=pred.get_reservation_id())
                        pr = self.actor.get_reservation(rid=predid)

                        if pr is None:
                            self.logger.warning(f"Predecessor for rid={res.get_reservation_id()} with rid={predid} "
                                                f"does not exist. Ignoring it!")
                            continue

                        if not isinstance(pr, ABCControllerReservation):
                            self.logger.warning(f"Predecessor for rid={res.get_reservation_id()} is not an "
                                                f"IControllerReservation: class={type(pr)}")
                            continue

                        self.logger.debug(f"Setting redeem predecessor on reservation # {res.get_reservation_id()} "
                                          f"pred={pr.get_reservation_id()}")
                        res.add_redeem_predecessor(reservation=pr)

                def run(self):
                    result = ResultAvro()
                    rid = ID(uid=reservation.get_reservation_id())
                    r = self.actor.get_reservation(rid=rid)
                    if r is None:
                        result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
                        result.set_message(ErrorCodes.ErrorNoSuchReservation.interpret())
                        return result

                    ManagementUtils.update_reservation(res_obj=r, rsv_mng=reservation)
                    if isinstance(reservation, LeaseReservationAvro):
                        predecessors = reservation.get_redeem_predecessors()
                        if predecessors is not None:
                            self.logger.debug("Processing Redeem predecessors")
                            self.__add_predecessors(res=r, predecessors=predecessors)

                    try:
                        self.actor.get_plugin().get_database().update_reservation(reservation=r)
                    except Exception as e:
                        self.logger.error("Could not commit slice update {}".format(e))
                        result.set_code(ErrorCodes.ErrorDatabaseError.value)
                        result.set_message(ErrorCodes.ErrorDatabaseError.interpret(exception=e))

                    self.actor.demand(rid=rid)

                    return result

            self.client.execute_on_actor_thread(runnable=Runner(actor=self.client, logger=self.logger))
        except Exception as e:
            self.logger.error("demand_reservation {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def extend_reservation(self, *, reservation: ID, new_end_time: datetime, sliver: BaseSliver,
                           caller: AuthToken, dependencies: List[ReservationPredecessorAvro] = None) -> ResultAvro:
        result = ResultAvro()

        if reservation is None or caller is None or (new_end_time is None and sliver is None):
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:
            class Runner(ABCActorRunnable):
                def __init__(self, *, actor: ABCActorMixin):
                    self.actor = actor

                def run(self):
                    result = ResultAvro()
                    r = self.actor.get_reservation(rid=reservation)
                    if r is None:
                        result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
                        result.set_message(ErrorCodes.ErrorNoSuchReservation.interpret())
                        return result

                    redeem_dep_res_list = []
                    if dependencies is not None:
                        for d in dependencies:
                            dep_res = self.actor.get_reservation(rid=ID(uid=d.get_reservation_id()))
                            if dep_res is None:
                                result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
                                result.set_message(ErrorCodes.ErrorNoSuchReservation.interpret())
                                return result
                            redeem_dep_res_list.append(dep_res)

                    rset = ResourceSet()
                    units = r.get_resources().get_units()
                    rset.set_units(units=units)
                    rset.set_type(rtype=r.get_resources().get_type())

                    new_term = r.get_term()
                    if new_end_time is not None:
                        tmp_start_time = r.get_term().get_start_time()
                        new_term = r.get_term().extend()

                        new_term.set_end_time(date=new_end_time)
                        new_term.set_new_start_time(date=tmp_start_time)
                        new_term.set_start_time(date=tmp_start_time)
                    if sliver is not None:
                        rset.set_sliver(sliver=sliver)

                    self.actor.extend(rid=r.get_reservation_id(), resources=rset, term=new_term,
                                      dependencies=redeem_dep_res_list)

                    return result
            '''
            # Process Extend for Renew synchronously
            if new_end_time is not None:
                result = self.client.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.client))
            # Process Extend for Modify asynchronously
            else:
                self.client.execute_on_actor_thread(runnable=Runner(actor=self.client))
            '''
            # Always Process Extend asynchronously
            self.client.execute_on_actor_thread(runnable=Runner(actor=self.client))

        except Exception as e:
            self.logger.error("extend_reservation {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def modify_reservation(self, *, rid: ID, modified_sliver: BaseSliver, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()

        if rid is None or modified_sliver is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        self.logger.debug("reservation: {} | modified_sliver= {}".format(rid, modified_sliver))
        try:

            class Runner(ABCActorRunnable):
                def __init__(self, *, actor: ABCClientActor):
                    self.actor = actor

                def run(self):
                    result = ResultAvro()
                    r = self.actor.get_reservation(rid=rid)
                    if r is None:
                        result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
                        result.set_message(ErrorCodes.ErrorNoSuchReservation.interpret())
                        return result

                    self.actor.modify(reservation_id=rid, modified_sliver=modified_sliver)

                    return result
            result = self.client.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.client))
        except Exception as e:
            self.logger.error("modify_reservation {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def claim_delegations(self, *, broker: ID, did: str, caller: AuthToken) -> ResultDelegationAvro:
        result = ResultDelegationAvro()
        result.status = ResultAvro()

        if caller is None or did is None or broker is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:
            my_broker = self.client.get_broker(guid=broker)

            if my_broker is None:
                result.status.set_code(ErrorCodes.ErrorNoSuchBroker.value)
                result.status.set_message(ErrorCodes.ErrorNoSuchBroker.interpret())
                return result

            class Runner(ABCActorRunnable):
                def __init__(self, *, actor: ABCClientActor):
                    self.actor = actor

                def run(self):
                    return self.actor.claim_delegation_client(delegation_id=did, broker=my_broker)

            rc = self.client.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.client))

            if rc is not None:
                result.delegations = []
                delegation = Translate.translate_delegation_to_avro(delegation=rc)
                result.delegations.append(delegation)
            else:
                raise ManageException("Internal Error")
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("claim_delegations {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def reclaim_delegations(self, *, broker: ID, did: str, caller: AuthToken) -> ResultDelegationAvro:
        result = ResultDelegationAvro()
        result.status = ResultAvro()

        if caller is None or did is None or broker is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        try:

            my_broker = self.client.get_broker(guid=broker)

            if my_broker is None:
                result.status.set_code(ErrorCodes.ErrorNoSuchBroker.value)
                result.status.set_message(ErrorCodes.ErrorNoSuchBroker.interpret())
                return result

            class Runner(ABCActorRunnable):
                def __init__(self, *, actor: ABCClientActor):
                    self.actor = actor

                def run(self):
                    return self.actor.reclaim_delegation_client(delegation_id=did, broker=my_broker)

            rc = self.client.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.client))

            if rc is not None:
                result.delegations = []
                delegation = Translate.translate_delegation_to_avro(delegation=rc)
                result.delegations.append(delegation)
            else:
                raise ManageException("Internal Error")
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("reclaim_delegations {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def poa(self, *, poa: PoaAvro, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()

        if poa.rid is None or poa.operation is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result

        self.logger.debug(f"reservation: {poa.rid} | operation = {poa.operation} | "
                          f"vcpu_cpu_map = {poa.vcpu_cpu_map} | node_set = {poa.node_set}")
        try:

            class Runner(ABCActorRunnable):
                def __init__(self, *, actor: ABCClientActor):
                    self.actor = actor

                def run(self):
                    result = ResultAvro()
                    r = self.actor.get_reservation(rid=ID(uid=poa.rid))
                    if r is None:
                        result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
                        result.set_message(ErrorCodes.ErrorNoSuchReservation.interpret())
                        return result

                    poa_obj = Translate.translate_poa_avro_to_poa(poa_avro=poa)

                    self.actor.poa(poa=poa_obj)

                    return result
            result = self.client.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.client))
        except Exception as e:
            self.logger.error("poa {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import os
import signal
import time
import traceback

import connexion
import prometheus_client
import waitress

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.util.graceful_interrupt_handler import GracefulInterruptHandler
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.swagger_server import encoder


def main():

    try:
        from fabric_cf.actor.core.container.globals import Globals, GlobalsSingleton
        with GracefulInterruptHandler() as h:

            GlobalsSingleton.get().start(force_fresh=False)

            while not GlobalsSingleton.get().start_completed:
                time.sleep(0.001)

            from fabric_cf.orchestrator.core.orchestrator_kernel import OrchestratorKernelSingleton
            OrchestratorKernelSingleton.get()
            OrchestratorKernelSingleton.get().start_threads()

            runtime_config = GlobalsSingleton.get().get_config().get_runtime_config()

            # prometheus server
            prometheus_port = int(runtime_config.get(Constants.PROPERTY_CONF_PROMETHEUS_REST_PORT, None))
            prometheus_client.start_http_server(prometheus_port)

            rest_port_str = runtime_config.get(Constants.PROPERTY_CONF_CONTROLLER_REST_PORT, None)

            if rest_port_str is None:
                raise OrchestratorException("Invalid configuration rest port not specified")

            print("Starting REST")
            # start swagger
            app = connexion.App(__name__, specification_dir='swagger_server/swagger/')
            app.json = encoder.JSONEncoder
            app.add_api('swagger.yaml', arguments={'title': 'Fabric Orchestrator API'}, pythonic_params=True)

            # Start up the server to expose the metrics.
            waitress.serve(app, port=int(rest_port_str), threads=8)

            while True:
                time.sleep(0.0001)
                if h.interrupted:
                    GlobalsSingleton.get().stop()
                    OrchestratorKernelSingleton.get().stop_threads()
    except Exception:
        traceback.print_exc()


if __name__ == '__main__':
    main()



from fabric_cf.orchestrator.swagger_server.models.resources import Resources  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response import resources_controller as rc


def portalresources_get(graph_format, level=None, force_refresh=None, start_date=None, end_date=None, includes=None, excludes=None):  # noqa: E501
    """Retrieve a listing and description of available resources for portal

    Retrieve a listing and description of available resources for portal # noqa: E501

    :param graph_format: graph format
    :type graph_format: str
    :param level: Level of details
    :type level: int
    :param force_refresh: Force to retrieve current available resource information.
    :type force_refresh: bool
    :param start_date: starting date to check availability from
    :type start_date: str
    :param end_date: end date to check availability until
    :type end_date: str
    :param includes: comma separated lists of sites to include
    :type includes: str
    :param excludes: comma separated lists of sites to exclude
    :type excludes: str

    :rtype: Resources
    """
    return rc.portalresources_get(graph_format=graph_format, level=level, force_refresh=force_refresh,
                                  start_date=start_date, end_date=end_date, includes=includes, excludes=excludes)


def resources_get(level, force_refresh, start_date=None, end_date=None, includes=None, excludes=None):  # noqa: E501
    """Retrieve a listing and description of available resources. By default, a cached available resource information is returned. User can force to request the current available resources.

    Retrieve a listing and description of available resources. By default, a cached available resource information is returned. User can force to request the current available resources. # noqa: E501

    :param level: Level of details
    :type level: int
    :param force_refresh: Force to retrieve current available resource information.
    :type force_refresh: bool
    :param start_date: starting date to check availability from
    :type start_date: str
    :param end_date: end date to check availability until
    :type end_date: str
    :param includes: comma separated lists of sites to include
    :type includes: str
    :param excludes: comma separated lists of sites to exclude
    :type excludes: str

    :rtype: Resources
    """
    return rc.resources_get(level=level, force_refresh=force_refresh, start_date=start_date,
                            end_date=end_date, includes=includes, excludes=excludes)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from abc import abstractmethod, ABC

from fim.slivers.base_sliver import BaseSliver

from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.resource_type import ResourceType


class ConfigToken(ABC):
    @abstractmethod
    def get_reservation_id(self) -> ID:
        """
        Return reservation id
        @return reservation
        """

    @abstractmethod
    def get_resource_type(self) -> ResourceType:
        """
        return resource type
        @return resource type
        """

    @abstractmethod
    def complete_modify(self):
        """
        Complete the modify operation
        """

    @abstractmethod
    def complete_poa(self, poa_info: dict):
        """
        Complete the poa operation
        """

    @abstractmethod
    def get_sequence(self) -> int:
        """
        Get Sequence number
        @return sequence number
        """

    @abstractmethod
    def get_id(self) -> ID:
        """
        Get unit id
        @return unit id
        """

    @abstractmethod
    def close(self):
        """
        Mark the unit as closed
        """

    @abstractmethod
    def activate(self):
        """
        Mark the unit as active
        """

    @abstractmethod
    def get_sliver(self) -> BaseSliver:
        """
        Get Sliver
        """

    @abstractmethod
    def set_sliver(self, *, sliver: BaseSliver):
        """
        Set sliver
        """

    @abstractmethod
    def update_sliver(self, *, sliver:BaseSliver):
        """
        Update Sliver
        """

    @abstractmethod
    def get_modified(self) -> BaseSliver:
        """
        Return the modified sliver
        """

    @abstractmethod
    def get_properties(self) -> dict:
        """
        Return properties
        """


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, List

from fim.graph.abc_property_graph import ABCPropertyGraph

from fabric_mb.message_bus.messages.reservation_mng import ReservationMng

from fabric_cf.actor.core.apis.abc_reservation_mixin import ReservationCategory
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.manage.server_actor_management_object import ServerActorManagementObject
from fabric_cf.actor.core.apis.abc_mgmt_server_actor import ABCMgmtServerActor
from fabric_cf.actor.core.manage.local.local_actor import LocalActor
from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.slice_avro import SliceAvro

    from fabric_cf.actor.core.manage.management_object import ManagementObject
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.manage.messages.client_mng import ClientMng
    from fabric_cf.actor.core.apis.abc_mgmt_actor import ABCMgmtActor


class LocalServerActor(LocalActor, ABCMgmtServerActor):
    def __init__(self, *, manager: ManagementObject, auth: AuthToken):
        super().__init__(manager=manager, auth=auth)

        if not isinstance(manager, ServerActorManagementObject):
            raise ManageException("Invalid manager object. Required: {}".format(type(ServerActorManagementObject)))

    def get_client_slices(self, *, id_token: str = None) -> List[SliceAvro]:
        self.clear_last()
        try:
            result = self.manager.get_client_slices(caller=self.auth)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.slices
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return None

    def get_clients(self, *, guid: ID = None, id_token: str = None) -> List[ClientMng]:
        self.clear_last()
        try:
            result = self.manager.get_clients(caller=self.auth, guid=guid)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.result
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return None

    def register_client(self, *, client: ClientMng, kafka_topic: str) -> bool:
        self.clear_last()
        if client is None or kafka_topic is None:
            self.last_exception = Exception("Invalid arguments")
            return False

        try:
            result = self.manager.register_client(client=client, kafka_topic=kafka_topic, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def update_client(self, *, client: ClientMng, kafka_topic: str) -> bool:
        self.clear_last()
        if client is None or kafka_topic is None:
            self.last_exception = Exception("Invalid arguments")
            return False

        try:
            result = self.manager.update_client(client=client, kafka_topic=kafka_topic, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def unregister_client(self, *, guid: ID) -> bool:
        self.clear_last()
        if guid is None:
            self.last_exception = Exception("Invalid arguments")
            return False

        try:
            result = self.manager.unregister_client(guid=guid, caller=self.auth)
            self.last_status = result

            return result.get_code() == 0
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return False

    def get_client_reservations(self, *, slice_id: ID = None, id_token: str = None) -> List[ReservationMng]:
        self.clear_last()
        try:
            result = self.manager.get_reservations_by_category(caller=self.auth,
                                                               category=ReservationCategory.Client,
                                                               slice_id=slice_id)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.reservations
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return None

    def get_broker_reservations(self, *, id_token: str = None) -> List[ReservationMng]:
        self.clear_last()
        try:
            result = self.manager.get_reservations_by_category(caller=self.auth,
                                                               category=ReservationCategory.Broker)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.reservations
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return None

    def get_inventory_slices(self, *, id_token: str = None) -> List[SliceAvro]:
        self.clear_last()
        try:
            result = self.manager.get_inventory_slices(caller=self.auth)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.slices
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return None

    def get_inventory_reservations(self, *, slice_id: ID = None, id_token: str = None) -> List[ReservationMng]:
        self.clear_last()
        try:
            result = self.manager.get_reservations_by_category(caller=self.auth,
                                                               category=ReservationCategory.Inventory,
                                                               slice_id=slice_id)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.reservations
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return None

    def add_client_slice(self, *, slice_mng: SliceAvro) -> ID:
        self.clear_last()
        try:
            result = self.manager.add_client_slice(caller=self.auth, slice_mng=slice_mng)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.get_result()
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

        return None

    def advertise_resources(self, *, delegation: ABCPropertyGraph, delegation_name: str, client: AuthToken) -> ID:
        try:
            result = self.manager.advertise_resources(delegation=delegation, delegation_name=delegation_name,
                                                      client=client, caller=self.auth)
            self.last_status = result.status
            if self.last_status.get_code() == 0 and result.get_result() is not None:
                return ID(uid=result.get_result())
        except Exception as e:
            self.on_exception(e=e, traceback_str=traceback.format_exc())

    def clone(self) -> ABCMgmtActor:
        return LocalServerActor(manager=self.manager, auth=self.auth)


# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server.models.status200_ok_single import Status200OkSingle  # noqa: F401,E501
from fabric_cf.orchestrator.swagger_server.models.version_data import VersionData  # noqa: F401,E501
from fabric_cf.orchestrator.swagger_server import util


class Version(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, size: int=1, status: int=200, type: str=None, data: List[VersionData]=None):  # noqa: E501
        """Version - a model defined in Swagger

        :param size: The size of this Version.  # noqa: E501
        :type size: int
        :param status: The status of this Version.  # noqa: E501
        :type status: int
        :param type: The type of this Version.  # noqa: E501
        :type type: str
        :param data: The data of this Version.  # noqa: E501
        :type data: List[VersionData]
        """
        self.swagger_types = {
            'size': int,
            'status': int,
            'type': str,
            'data': List[VersionData]
        }

        self.attribute_map = {
            'size': 'size',
            'status': 'status',
            'type': 'type',
            'data': 'data'
        }
        self._size = size
        self._status = status
        self._type = type
        self._data = data

    @classmethod
    def from_dict(cls, dikt) -> 'Version':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The version of this Version.  # noqa: E501
        :rtype: Version
        """
        return util.deserialize_model(dikt, cls)

    @property
    def size(self) -> int:
        """Gets the size of this Version.


        :return: The size of this Version.
        :rtype: int
        """
        return self._size

    @size.setter
    def size(self, size: int):
        """Sets the size of this Version.


        :param size: The size of this Version.
        :type size: int
        """

        self._size = size

    @property
    def status(self) -> int:
        """Gets the status of this Version.


        :return: The status of this Version.
        :rtype: int
        """
        return self._status

    @status.setter
    def status(self, status: int):
        """Sets the status of this Version.


        :param status: The status of this Version.
        :type status: int
        """

        self._status = status

    @property
    def type(self) -> str:
        """Gets the type of this Version.


        :return: The type of this Version.
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type: str):
        """Sets the type of this Version.


        :param type: The type of this Version.
        :type type: str
        """

        self._type = type

    @property
    def data(self) -> List[VersionData]:
        """Gets the data of this Version.


        :return: The data of this Version.
        :rtype: List[VersionData]
        """
        return self._data

    @data.setter
    def data(self, data: List[VersionData]):
        """Sets the data of this Version.


        :param data: The data of this Version.
        :type data: List[VersionData]
        """

        self._data = data
